<?php

include_once './constants.php';

/**
 * compiles the given main file (if a compile error occurs then an answer is sent)
 * @param string $mainFileNameWithExtension the main file to compile (with extension)
 * @param $fullWorkingDirPath string the working dir with the source files
 * @param string $compileCmd the compile command to execute
 * @param $sourceFileExtensions array the list with the file extensions from the source files for the p language
 * @param $uniqueSessionId string the unique session id for the test runner to handle sandbox file system management
 * @param $showTestRunnerDebugOutput true: show internal debug output from the test runner (included in the test response)
 * @return array
 * Format: return_var: process exit value, output the output of the process
 * @internal param array $allFiles all files to copy in teh dir to use
 * Format: array[int] => {fileName: string, fileContent: string}
 */
function do_compile($mainFileNameWithExtension, $fullWorkingDirPath, $compileCmd, $sourceFileExtensions,
                    $uniqueSessionId,
                    $showTestRunnerDebugOutput
)
{

    global $config;
    global $s_output;
    global $s_return_val;
    global $s_command_compile;

    $command_to_execute_string = $s_command_compile;

    $output = [];
    $return_var = -1;

    $longCmd = $config['runner']
        . ' ' . $command_to_execute_string                      # arg[0] the test method
        . ' "' . $fullWorkingDirPath . '"'                      # arg[1] dir path
        . ' ' . $mainFileNameWithExtension                      # arg[2] file path
        . ' "' . $compileCmd . '"'                              # arg[3] command to execute
        . ' "' . implode(',', $sourceFileExtensions) . '"'# arg[4] the source file extensions
        . ' "' . $uniqueSessionId . '"'                         # arg[5] a unique session id
        . ' "' . $showTestRunnerDebugOutput . '"'              # arg[6] showTestRunnerDebugOutput
    ;

    exec($longCmd, $output, $return_var);

    /* could also be used...proc_open only works if bypass_shell is true??
    $descriptorspec = array(
        0 => array('pipe', 'r'), # stdin is a pipe that the child will read from
        1 => array('pipe', 'w'),  # stdout is a pipe that the child will write to
        2 => array("pipe", "w")
    );

    $process = proc_open($longCmd, $descriptorspec, $pipes, NULL, NULL, array('bypass_shell' => TRUE));

    if (is_resource($process)) {

        $output = stream_get_contents($pipes[1]);
        fclose($pipes[1]);

        $errorOutput = stream_get_contents($pipes[2]);
        fclose($pipes[2]);

        $return_var = proc_close($process);
    }
    */

    return array(
        $s_return_val => $return_var,
        $s_output => $output
    );
}
