<?php


/**
 * @deprecated with json we need to escape the regex strings... which is messy
 * format2 is similar to json but without escapes --> use do_regexTest_format2
 * runs the given regex test against the given files
 * @param $allFiles array the files
 * Format: array[int] => {fileName: string, fileContent: string}
 * @param $testContent string the test content
 * @param $showTestRunnerDebugOutput true: show internal debug output from the test runner (included in the test response)
 * @return array
 */
function do_regexTest_json($allFiles, $testContent, $showTestRunnerDebugOutput)
{

    global $availableRegexModes;
    global $s_regexMode;
    global $regexModeBlacklist;
    global $regexModeWhitelist;
    global $s_regexConnector;
    global $availableRegexConnectors;
    global $regexAndConnector;
    global $regexOrConnector;
    global $regexMathAllFilesAttribute;

    global $s_return_val;
    global $s_output;
    global $s_fileName;
    global $s_fileContent;
    global $config;

    # removes leading and trailing whitespaces ??
    $unformattedTestContent = trim(preg_replace('/\s\s+/', ' ', $testContent));

    $testObj = json_decode($unformattedTestContent, TRUE);

    $output = [];
    $return_var = 0;

    if ($testObj === NULL) {
        $return_var = 50;
        array_push($output, "could not decode regex test content (json obj) (maybe some snytax error or a property key is not unique)");

        goto handleOutputLabel;
    }

    $mode = $regexModeBlacklist;
    $connector = $regexAndConnector;


    //check for regex mode
    if (isset($testObj[$s_regexMode])) {
        $mode = $testObj[$s_regexMode];

        if (!in_array($mode, $availableRegexModes)) {
            $return_var = 50;
            array_push($output, "invalid regex mode");

            goto handleOutputLabel;
        }
    }

    //check for regex connector
    if (isset($testObj[$s_regexConnector])) {
        $connector = $testObj[$s_regexConnector];

        if (!in_array($connector, $availableRegexConnectors)) {
            $return_var = 50;
            array_push($output, "invalid regex connector");

            goto handleOutputLabel;
        }
    }

    # use the neutral element for the initial value for the array_reduce (should not change the result)
    $neutralElement = NULL;

    if ($connector === $regexAndConnector) {
        $neutralElement = TRUE;

    } else /*if ($connect === $regexOrConnector) */ {
        $neutralElement = FALSE;
    }

    array_push($output, 'used mode: ' . $mode);
    array_push($output, 'used connector: ' . $connector);

    $resultVector = [];

    foreach ($testObj as $key => $item) {

        if ($key === $s_regexMode || $key === $s_regexConnector) {
            continue;
        }

        # every property is a regex for a file

        $file = $key;
        $fullRegex = $item;

        # find the right file...

        # the $resultVector contains blacklisted/whitelisted results
        #the blacklist inverts the results because files:
        # a -> true (regex found) -> false
        # b -> true (regex found) -> false
        # c -> false (regex not found) -> true (because regex should not be found)

        if ($key === $regexMathAllFilesAttribute) {

            //run the regex against all files not just one... because user requested it (with *)
            foreach ($allFiles as $fileTuple) {
                # run the regex test for this file
                if (_regexTest_helper($mode, $connector, $fullRegex, $fileTuple, $resultVector, $return_var, $output) === FALSE) {

                    //status code and output is set in the _regexTest_helper method
                    goto handleOutputLabel;
                }
            }

        } else {

            //run the regex against the matching file
            $fileFound = FALSE;
            foreach ($allFiles as $fileTuple) {
                if ($fileTuple[$s_fileName] === $file) {
                    # run the regex test for this file
                    if (_regexTest_helper($mode, $connector, $fullRegex, $fileTuple, $resultVector, $return_var, $output) === FALSE) {

                        //status code and output is set in the _regexTest_helper method
                        goto handleOutputLabel;
                    }
                    $fileFound = TRUE;
                    break;
                }
            }

            if ($fileFound === FALSE) {
                //file was not found .... result is always "true" true for blacklist and false for whitelist
                //common true = $neutralElement

                //if we do not find the file then it is not added to the output (_regexTest_helper adds the output)
                //this is useful when we create a regex test for multiple languages (--> different files + extensions...)
                if (isset($resultVector[$file])) {

                    if ($connector === $regexAndConnector) {
                        $resultVector[$file] = $resultVector[$file] && $neutralElement;
                    }
                     else {// if ($connector === $regexOrConnector) {
                         $resultVector[$file] = $resultVector[$file] || $neutralElement;
                     }

                }  else {
                    $resultVector[$file] = $neutralElement;
                }
            }
        }
    }

    # after collecting all results merge together

    # use the neutral element for the initial value for the array_reduce (should not change the result)



    if (count($resultVector) === 0) {

        $return_var = 2;
        array_push($output, "no test results found");

        goto handleOutputLabel;
    }

    $fullRegexTestResult = array_reduce($resultVector, function ($carry, $current) use ($mode, $regexModeBlacklist, $connector, $regexAndConnector, $regexOrConnector) {

        if ($connector === $regexAndConnector) {
            return $carry && $current;
        } else {
            return $carry || $current;
        }

    }, $neutralElement);

    if ($fullRegexTestResult === NULL) {

        $return_var = 2;
        array_push($output, "error getting regex test result");

        goto handleOutputLabel;
    }

    //test ran, 0 = success, 0 = fail
    $return_var = $fullRegexTestResult === TRUE ? 0 : 1;
    //output for every file the result...

    handleOutputLabel:

    return array(
        $s_return_val => $return_var,
        $s_output => $output
    );

    #handleExecOutput($output, $return_var, $config, $conn, $rootDirName, $transactionKey);
}

/**
 * runs the given regex test against the given files
 * @param $allFiles array the files
 * Format: array[int] => {fileName: string, fileContent: string}
 * @param $testContent string the test content
 * @return array
 */
function do_regexTest_format2($allFiles, $testContent)
{

    global $availableRegexModes;
    global $s_regexMode;
    global $regexModeBlacklist;
    global $regexModeWhitelist;
    global $s_regexConnector;
    global $availableRegexConnectors;
    global $regexAndConnector;
    global $regexOrConnector;
    global $regexMathAllFilesAttribute;

    global $s_return_val;
    global $s_output;
    global $s_fileName;
    global $s_fileContent;
    global $config;

    # removes leading and trailing whitespaces ??
    $unformattedTestContent = $testContent; //trim(preg_replace('/\s\s+/', ' ', $testContent));

    if ($unformattedTestContent === NULL || $unformattedTestContent === '') {
        $return_var = 50;
        array_push($output, "could not decode regex test (no content)");

        goto handleOutputLabel;
    }

    $lines = explode("\n", $unformattedTestContent);

    $output = [];
    $return_var = 0;

    if (count($lines) === 0) {
        $return_var = 50;
        array_push($output, "could not decode regex test (no lines)");

        goto handleOutputLabel;
    }

    # associative array of {fileName => regex}
    $testObj = array();

    $lineCount = 1;
    foreach ($lines as $line) {

        if (trim($line) === "") {
            continue;
        }

        if (substr(trim($line), 0, 1) === "#") { //comment --> ignore
            continue;
        }

        $colonPos = strpos($line, ':');

        if ($colonPos === FALSE) {
            $return_var = 50;
            array_push($output, "could not decode regex test (no : found), error on line " . $lineCount);
            goto handleOutputLabel;
        }

        $fileName = trim(substr($line, 0, $colonPos));
        $regex = trim(substr($line, $colonPos+1));

        if ($fileName === "") {
            $return_var = 50;
            array_push($output, "could not find file name, error on line " . $lineCount);
            goto handleOutputLabel;
        }

        if ($regex === "") {
            $return_var = 50;
            array_push($output, "could not find regex, error on line " . $lineCount);
            goto handleOutputLabel;
        }

        if (isset($testObj[$fileName])) {
            $return_var = 50;
            array_push($output, "duplicate file name: " . $fileName);
            goto handleOutputLabel;
        }

        $testObj[$fileName] = $regex;
        $lineCount++;
    }


    $mode = $regexModeBlacklist;
    $connector = $regexAndConnector;


    //check for regex mode
    if (isset($testObj[$s_regexMode])) {
        $mode = $testObj[$s_regexMode];

        if (!in_array($mode, $availableRegexModes)) {
            $return_var = 50;
            array_push($output, "invalid regex mode");

            goto handleOutputLabel;
        }
    }

    //check for regex connector
    if (isset($testObj[$s_regexConnector])) {
        $connector = $testObj[$s_regexConnector];

        if (!in_array($connector, $availableRegexConnectors)) {
            $return_var = 50;
            array_push($output, "invalid regex connector");

            goto handleOutputLabel;
        }
    }

    # use the neutral element for the initial value for the array_reduce (should not change the result)
    $neutralElement = NULL;

    if ($connector === $regexAndConnector) {
        $neutralElement = TRUE;

    } else /*if ($connect === $regexOrConnector) */ {
        $neutralElement = FALSE;
    }

    array_push($output, 'used mode: ' . $mode);
    array_push($output, 'used connector: ' . $connector);

    $resultVector = [];

    foreach ($testObj as $key => $item) {

        if ($key === $s_regexMode || $key === $s_regexConnector) {
            continue;
        }

        # every property is a regex for a file

        $file = $key;
        $fullRegex = $item;

        # find the right file...

        # the $resultVector contains blacklisted/whitelisted results
        #the blacklist inverts the results because files:
        # a -> true (regex found) -> false
        # b -> true (regex found) -> false
        # c -> false (regex not found) -> true (because regex should not be found)

        if ($key === $regexMathAllFilesAttribute) {

            //run the regex against all files not just one... because user requested it (with *)
            foreach ($allFiles as $fileTuple) {
                # run the regex test for this file
                if (_regexTest_helper($mode, $connector, $fullRegex, $fileTuple, $resultVector, $return_var, $output) === FALSE) {

                    //status code and output is set in the _regexTest_helper method
                    goto handleOutputLabel;
                }
            }

        } else {

            //run the regex against the matching file
            $fileFound = FALSE;
            foreach ($allFiles as $fileTuple) {
                if ($fileTuple[$s_fileName] === $file) {
                    # run the regex test for this file
                    if (_regexTest_helper($mode, $connector, $fullRegex, $fileTuple, $resultVector, $return_var, $output) === FALSE) {

                        //status code and output is set in the _regexTest_helper method
                        goto handleOutputLabel;
                    }
                    $fileFound = TRUE;
                    break;
                }
            }

            if ($fileFound === FALSE) {
                //file was not found .... result is always "true" true for blacklist and false for whitelist
                //common true = $neutralElement

                //if we do not find the file then it is not added to the output (_regexTest_helper adds the output)
                //this is useful when we create a regex test for multiple languages (--> different files + extensions...)
                if (isset($resultVector[$file])) {

                    if ($connector === $regexAndConnector) {
                        $resultVector[$file] = $resultVector[$file] && $neutralElement;
                    }
                    else {// if ($connector === $regexOrConnector) {
                        $resultVector[$file] = $resultVector[$file] || $neutralElement;
                    }

                }  else {
                    $resultVector[$file] = $neutralElement;
                }
            }
        }
    }

    # after collecting all results merge together

    # use the neutral element for the initial value for the array_reduce (should not change the result)



    if (count($resultVector) === 0) {

        $return_var = 2;
        array_push($output, "no test results found");

        goto handleOutputLabel;
    }

    $fullRegexTestResult = array_reduce($resultVector, function ($carry, $current) use ($mode, $regexModeBlacklist, $connector, $regexAndConnector, $regexOrConnector) {

        if ($connector === $regexAndConnector) {
            return $carry && $current;
        } else {
            return $carry || $current;
        }

    }, $neutralElement);

    if ($fullRegexTestResult === NULL) {

        $return_var = 2;
        array_push($output, "error getting regex test result");

        goto handleOutputLabel;
    }

    //test ran, 0 = success, 0 = fail
    $return_var = $fullRegexTestResult === TRUE ? 0 : 1;
    //output for every file the result...

    handleOutputLabel:

    return array(
        $s_return_val => $return_var,
        $s_output => $output
    );

    #handleExecOutput($output, $return_var, $config, $conn, $rootDirName, $transactionKey);
}

/**
 * runs a regex test against the given file
 * @param string $mode the mode to use
 * @param string $connector used to connect two results from one file
 * @param string $fullRegex the regex to use
 * @param array $fileTuple the file and its content
 * @param array $resultVector the result vector (reference)
 * @param $return_var int the temp return var (code) in case some error occurred
 * @param array $output the response protocol
 * @return bool true: all ok, false: some error
 */
function _regexTest_helper($mode, $connector, $fullRegex, $fileTuple, &$resultVector, &$return_var, &$output)
{

    global $regexModeBlacklist;
    global $s_fileContent;
    global $s_fileName;
    global $regexAndConnector;
    global $regexOrConnector;
    global $regexAndInlineConnector;
    global $regexOrInlineConnector;

    $fileName = $fileTuple[$s_fileName];
    $fileContent = $fileTuple[$s_fileContent];

    #$fullRegex can contain multiple regex expressions that are connected through && or/and ||

    $allRegexExpressions = _getRegexExpressions($fullRegex, $return_var, $output);

    if ($allRegexExpressions === NULL) {
        //exit(1); //if we exit then the cleanup won't be done!
        return false;
    }

    $result = NULL;
    $nextConnector = '';



    # $allRegexExpressions should contain a regex, a connector, a regex, a connector, a regex, ...
    foreach ($allRegexExpressions as $regexExpression) {

        if ($regexExpression === $regexAndInlineConnector) {
            $nextConnector = $regexAndInlineConnector;
            continue;
        } else if ($regexExpression === $regexOrInlineConnector) {
            $nextConnector = $regexOrInlineConnector;
            continue;
        }

        if ($result === NULL) {
            # first result
            $regexResult = preg_match($regexExpression, $fileContent);

            if ($regexResult === FALSE) {

                //try to get the last error
                $lastRegexError = preg_last_error();
                $errorText = _getLastRegexErrorText($lastRegexError);

                $return_var = 2;
                array_push($output, 'could not execute regex: ' . $regexExpression . ', error: ' . $errorText);

                return false;
                //exit(1); //if we exit then the cleanup won't be done!
            }

            $result = $regexResult === 1 ? TRUE : FALSE;

            if ($mode === $regexModeBlacklist) {
                //inverted result see below for explanation
                $result = !$result;
            }
            array_push($output, $fileName . ' -> (' . $regexExpression . ') -> ' . var_export($result, TRUE));

        } else {

            $regexResult = preg_match($regexExpression, $fileContent);

            if ($regexResult === FALSE) {

                //try to get the last error
                $lastRegexError = preg_last_error();
                $errorText = _getLastRegexErrorText($lastRegexError);

                $return_var = 2;
                array_push($output, 'could not execute regex: ' . $regexExpression . ', error: ' . $errorText);

                return false;
                //exit(1); //if we exit then the cleanup won't be done!
            }
            $temp = $regexResult === 1 ? TRUE : FALSE;

            if ($mode === $regexModeBlacklist) {
                //inverted result see below for explanation
                $temp = !$temp;
            }
            array_push($output, $fileName . ' -> (' . $regexExpression . ') -> ' . var_export($temp, TRUE));

            if ($nextConnector === $regexAndInlineConnector) {
                $result = $result && $temp;
            } else if ($nextConnector === $regexOrInlineConnector) {
                $result = $result || $temp;
            }
        }
    }

    #$result = preg_match($fullRegex, $fileContent);

    # 1 = found, 0 = not found
    $newResult = $result === NULL ? TRUE : $result; //if NULL then there was no regex --> success

    # full result if we have more than 1 regex
    if (count($allRegexExpressions) > 1) {
        array_push($output, $fileName . ' (full regex) -> (' . $fullRegex . ') -> ' . var_export($newResult, TRUE));
    }

    # we can have multiple results for one file... merge if needed

    if (isset($resultVector[$fileName])) {
        $oldValue = $resultVector[$fileName];

        if ($connector === $regexAndConnector) {
            $resultVector[$fileName] = $oldValue && $newResult;

        } else /*if ($connect === $regexOrConnector) */ {
            $resultVector[$fileName] = $oldValue || $newResult;
        }

    } else {
        $resultVector[$fileName] = $newResult;
    }

    return TRUE;
}

/**
 * @param $fullRegexLine
 * @param $return_var int the temp return var (code) in case some error occurred
 * @param array $output the response protocol
 * @return array an array containing 2n elements, 1. is the regex 2. is the connector for the next regex result
 * OR NULL on error
 *(the last connector must be ignored)
 */
function _getRegexExpressions($fullRegexLine, &$return_var, &$output)
{

    global $regexAndInlineConnector;
    global $regexOrInlineConnector;

    #$is_regex_regex = "/^\/[\s\S]+\/$/";
    $testString = 'test';

    $resultExpressions = [];

    //if no && is contained then explode returns the original expression
    $andConnectedRegexExpr = explode($regexAndInlineConnector, $fullRegexLine);

    //if we have at least one && then we have at least 2 expressions
    if (count($andConnectedRegexExpr) > 1) {

        foreach ($andConnectedRegexExpr as $andRegex) {
            $orConnectedRegexExpr = explode($regexOrInlineConnector, $andRegex);

            # check if $orConnectedRegexExpr is really a regex (a basic check)

            if (count($orConnectedRegexExpr) === 1) {

                if (preg_match($orConnectedRegexExpr[0], $testString) === FALSE) { // FALSE = error -> invalid regex
                    $return_var = 2;
                    array_push($output, $orConnectedRegexExpr . ' is not a regex');
                    return NULL;
                }

                # in this case the $regexOrInlineConnector did nothing and we just had 1 && regex
                array_push($resultExpressions, $orConnectedRegexExpr[0]); # push regex
                array_push($resultExpressions, $regexAndInlineConnector); # push connector to next regex
                continue;
            }

            foreach ($orConnectedRegexExpr as $atomRegex) {
                if (preg_match($atomRegex, $testString) === FALSE) {
                    $return_var = 2;
                    array_push($output, $atomRegex . ' is not a regex');
                    return NULL;
                }
                array_push($resultExpressions, $atomRegex); # push regex
                array_push($resultExpressions, $regexOrInlineConnector); # push connector to next regex
            }
        }
    } else {
        //if the regex just contains || ... || ... || then there is no and expression ...
        $orConnectedRegexExpr = explode($regexOrInlineConnector, $fullRegexLine);

        if (count($orConnectedRegexExpr) === 1) {
            //this can happen if the regex does not contain || or && expressions --> just 1 normal regex

            //check if it's really a regex
            if (preg_match($fullRegexLine, $testString) === FALSE) {
                $return_var = 2;
                array_push($output, $fullRegexLine . ' is not a regex');
                return NULL;
            }

            array_push($resultExpressions, $fullRegexLine); # push regex

        } else {

            foreach ($orConnectedRegexExpr as $atomRegex) {
                if (preg_match($atomRegex, $testString) === FALSE) {
                    $return_var = 2;
                    array_push($output, $atomRegex . ' is not a regex');
                    return NULL;
                }

                array_push($resultExpressions, $atomRegex); # push regex
                array_push($resultExpressions, $regexOrInlineConnector); # push connector to next regex
            }
        }

    }


    return $resultExpressions;
}

function _getLastRegexErrorText($lastRegexError) {
    $errorReason = '';
    if ($lastRegexError === PREG_NO_ERROR) {
        $errorReason = 'PREG_NO_ERROR this should not happen??';

    } else if ($lastRegexError === PREG_INTERNAL_ERROR) {
        $errorReason = 'PREG_INTERNAL_ERROR (some internal php error)';

    } else if ($lastRegexError === PREG_BACKTRACK_LIMIT_ERROR) {
        $errorReason = 'PREG_BACKTRACK_LIMIT_ERROR';

    } else if ($lastRegexError === PREG_RECURSION_LIMIT_ERROR) {
        $errorReason = 'PREG_RECURSION_LIMIT_ERROR';

    } else if ($lastRegexError === PREG_BAD_UTF8_ERROR) {
        $errorReason = 'PREG_BAD_UTF8_ERROR';

    } else if ($lastRegexError === PREG_BAD_UTF8_OFFSET_ERROR) {
        $errorReason = 'PREG_BAD_UTF8_OFFSET_ERROR';

    } else if ($lastRegexError === PREG_JIT_STACKLIMIT_ERROR) {
        $errorReason = 'PREG_JIT_STACKLIMIT_ERROR';
    } else {
        $errorReason = 'invalid error enum value';
        warn('invalid regex error enum value: ' . $lastRegexError);
    }

    $errorReason = $errorReason . ' [see http://php.net/manual/de/function.preg-last-error.php for more info]';

    return $errorReason;
}